# -*- coding: utf-8 -*-
"""
computes the tonal power ratio from the magnitude spectrum

  Args:
    X: spectrogram (dimension FFTLength X Observations)
    f_s: sample rate of audio data
    G_T: energy threshold

  Returns:
    vtpr tonal power ratio
"""

import numpy as np
from scipy.signal import find_peaks


def FeatureSpectralTonalPowerRatio(X, f_s, G_T=5e-4):

    isSpectrum = X.ndim == 1
    if isSpectrum:
        X = np.expand_dims(X, axis=1)

    X = X**2

    fSum = X.sum(axis=0)
    vtpr = np.zeros(fSum.shape)

    for n in range(0, X.shape[1]):
        if fSum[n] < G_T:
            continue

        # find local maxima above the threshold
        afPeaks = find_peaks(X[:, n], height=G_T)

        if not afPeaks[0].size:
            continue

        # calculate ratio
        vtpr[n] = X[afPeaks[0], n].sum() / fSum[n]

    return np.squeeze(vtpr) if isSpectrum else vtpr



# -*- coding: utf-8 -*-
"""
helper function: convert Midi pitch (floating point) from frequency

  Args:
    fInHz: The frequency to be converted, can be scalar or vector
    fA4InHz: The reference tuning frequency (default: 440Hz)

  Returns:
    Midi values of the input dimension (floating point)
"""

import numpy as np


def ToolFreq2Midi(fInHz, fA4InHz = 440):
    def convert_freq2midi_scalar(f, fA4InHz):
 
        if f <= 0:
            return 0
        else:
            return (69 + 12 * np.log2(f/fA4InHz))

    fInHz = np.asarray(fInHz)
    if fInHz.ndim == 0:
       return convert_freq2midi_scalar(fInHz,fA4InHz)

    midi = np.zeros(fInHz.shape)
    for k,f in enumerate(fInHz):
        midi[k] =  convert_freq2midi_scalar(f,fA4InHz)
            
    return (midi)




import numpy as np
from scipy.signal import lfilter


# see function mfcc.m from Slaneys Auditory Toolbox (Matlab)
def ToolGammatoneFb(afAudioData, f_s, iNumBands=20, f_low=100):

    # initialization
    fEarQ = 9.26449
    fBW = 24.7
    iOrder = 1
    T = 1 / f_s

    # allocate output memory
    X = np.zeros([iNumBands, afAudioData.shape[0]])

    # compute the mid frequencies
    f_c = getMidFrequencies(f_low, f_s / 2, iNumBands, fEarQ, fBW)

    # compute the coefficients
    [afCoeffB, afCoeffA] = getCoeffs(f_c, 1.019 * 2 * np.pi * (((f_c / fEarQ)**iOrder + fBW**iOrder)**(1 / iOrder)), T)

    # do the (cascaded) filter process
    for k in range(0, iNumBands):
        X[k, :] = afAudioData
        for j in range(0, 4):
            X[k, :] = lfilter(afCoeffB[j, :, k], afCoeffA[j, :, k], X[k, :])

    return (X)


# see function ERBSpace.m from Slaneys Auditory Toolbox
def getMidFrequencies(f_low, f_hi, iNumBands, fEarQ, fBW):

    freq = np.log((f_low + fEarQ * fBW) / (f_hi + fEarQ * fBW)) / iNumBands
    f_c = np.exp(np.arange(1, iNumBands + 1) * freq)
    f_c = -(fEarQ * fBW) + f_c * (f_hi + fEarQ * fBW)

    return (f_c)


# see function MakeERBFilters.m from Slaneys Auditory Toolbox
def getCoeffs(f_c, B, T):

    fCos = np.cos(2 * f_c * np.pi * T)
    fSin = np.sin(2 * f_c * np.pi * T)
    fExp = np.exp(B * T)
    fSqrtA = 2 * np.sqrt(3 + 2**(3 / 2))
    fSqrtS = 2 * np.sqrt(3 - 2**(3 / 2))

    A0 = T
    A2 = 0
    B0 = 1
    B1 = -2 * fCos / fExp
    B2 = np.exp(-2 * B * T)

    A11 = -(2 * T * fCos / fExp + fSqrtA * T * fSin / fExp) / 2
    A12 = -(2 * T * fCos / fExp - fSqrtA * T * fSin / fExp) / 2
    A13 = -(2 * T * fCos / fExp + fSqrtS * T * fSin / fExp) / 2
    A14 = -(2 * T * fCos / fExp - fSqrtS * T * fSin / fExp) / 2

    fSqrtA = np.sqrt(3 + 2**(3 / 2))
    fSqrtS = np.sqrt(3 - 2**(3 / 2))
    fArg = (f_c * np.pi * T) * 1j

    fExp1 = 2 * np.exp(4 * fArg)
    fExp2 = 2 * np.exp(-(B * T) + 2 * fArg)

    afGain = np.abs((-fExp1 * T + fExp2 * T * (fCos - fSqrtS * fSin)) *
                    (-fExp1 * T + fExp2 * T * (fCos + fSqrtS * fSin)) *
                    (-fExp1 * T + fExp2 * T * (fCos - fSqrtA * fSin)) *
                    (-fExp1 * T + fExp2 * T * (fCos + fSqrtA * fSin)) /
                    (-2 / np.exp(2 * B * T) - fExp1 + (2 + fExp1) / fExp)**4)

    # this is Slaney's compact format - now resort into 3D Matrices
    # fcoefs = [A0*ones(length(f_c),1) A11 A12 A13 A14 A2*ones(length(f_c),1) B0*ones(length(f_c),1) B1 B2 afGain];

    afCoeffB = np.zeros([4, 3, B.size])
    afCoeffA = np.zeros([4, 3, B.size])

    for k in range(0, B.size):
        afCoeffB[0, :, k] = [A0, A11[k], A2] / afGain[k]
        afCoeffA[0, :, k] = [B0, B1[k], B2[k]]

        afCoeffB[1, :, k] = [A0, A12[k], A2]
        afCoeffA[1, :, k] = [B0, B1[k], B2[k]]

        afCoeffB[2, :, k] = [A0, A13[k], A2]
        afCoeffA[2, :, k] = [B0, B1[k], B2[k]]

        afCoeffB[3, :, k] = [A0, A14[k], A2]
        afCoeffA[3, :, k] = [B0, B1[k], B2[k]]

    return (afCoeffB, afCoeffA)


#############################################################################
#    # initialization
#    f_start         = 133.3333
#
#    iNumLinFilters  = 13
#    iNumLogFilters  = 27
#    iNumFilters     = iNumLinFilters + iNumLogFilters
#
#    linearSpacing   = 66.66666666
#    logSpacing      = 1.0711703
#
#    # compute band frequencies
#    f = np.zeros(iNumFilters+2)
#    f[np.arange(0,iNumLinFilters)] = f_start + np.arange(0,iNumLinFilters)*linearSpacing
#    f[np.arange(iNumLinFilters,iNumFilters+2)] = f[iNumLinFilters-1] * logSpacing**np.arange(1,iNumLogFilters+3)
#
#    # sanity check
#    if f[iNumLinFilters-1]>=f_s/2:
#        f = f[f<f_s/2]
#        iNumFilters = f.shape[0] - 2
#
#    f_l = f[np.arange(0,iNumFilters)]
#    f_c = f[np.arange(1,iNumFilters+1)]
#    f_u = f[np.arange(2,iNumFilters+2)]
#
#    # allocate memory for filters and set max amplitude
#    H = np.zeros([iNumFilters,iFftLength])
#    afFilterMax = 2 / (f_u - f_l)
#    f_k = np.arange(0,iFftLength)/(iFftLength-1)*f_s/2
#
#    # compute the transfer functions
#    for c in range(0,iNumFilters):
#        #lower filter slope
#        i_l = np.argmax(f_k>f_l[c])
#        i_u = np.max([0, np.argmin(f_k <= f_c[c])-1])
#        H[c,np.arange(i_l, i_u+1)] = afFilterMax[c] * (f_k[np.arange(i_l, i_u+1)]-f_l[c])/(f_c[c]-f_l[c])
#        #upper filter slope
#        i_l = i_u + 1
#        i_u = np.max([0, np.argmin(f_k < f_u[c])-1])
#        H[c,np.arange(i_l, i_u+1)] = afFilterMax[c] * (f_u[c]-f_k[np.arange(i_l, i_u+1)])/(f_u[c]-f_c[c])
#
#    return (H)



# -*- coding: utf-8 -*-
"""
computes the ACF maxima of a time domain signal

  Args:
    x: audio signal
    iBlockLength: block length in samples
    iHopLength: hop length in samples
    f_s: sample rate of audio data (unused)

  Returns:
    vta autocorrelation maximum
    t time stamp

"""

import numpy as np
import pyACA


def FeatureTimeMaxAcf(x, iBlockLength, iHopLength, f_s, f_max=2000, fMinThresh=0.35):

    # create blocks
    xBlocks = pyACA.ToolBlockAudio(x, iBlockLength, iHopLength)

    # number of results
    iNumOfBlocks = xBlocks.shape[0]

    # compute time stamps
    t = (np.arange(0, iNumOfBlocks) * iHopLength + (iBlockLength / 2)) / f_s

    # allocate memory
    vacf = np.zeros(iNumOfBlocks)

    for n, block in enumerate(xBlocks):
        eta_min = np.floor(f_s / f_max).astype(int)

        # calculate the acf
        if not block.sum():
            continue
        else:
            afCorr = np.correlate(block, block, "full") / np.dot(block, block)

        afCorr = afCorr[np.arange(iBlockLength, afCorr.size)]

        # update eta_min to avoid main lobe
        eta_tmp = np.argmax(afCorr < fMinThresh)
        eta_min = np.max([eta_min, eta_tmp])

        afDeltaCorr = np.diff(afCorr)
        eta_tmp = np.argmax(afDeltaCorr > 0)
        eta_min = np.max([eta_min, eta_tmp])

        # find the coefficients specified in eta
        vacf[n] = np.max(afCorr[np.arange(eta_min + 1, afCorr.size)])

    return vacf, t



import unittest
import numpy as np
import numpy.testing as npt

import pyACA


class TestShape(unittest.TestCase):

    def test_compute_feature_num_blocks(self):
        blockLength = 256   # Using smaller blocksize to reduce time required to test (~7.3 sec)
        hopLength = 128
        fs = 44100
        inputData = np.random.uniform(-1, 1, size=(2*blockLength, 1))

        for inputSize in range(1, 2*blockLength):
            x = inputData[:inputSize]
            expectedNumBlocks = self.calcNumBlocks(x.size+blockLength, blockLength, hopLength)
            for feature in pyACA.getFeatureList('all'):
                with self.subTest(msg=feature + ':' + str(inputSize)):
                    out, t = pyACA.computeFeature(feature, x, fs, iBlockLength=blockLength, iHopLength=hopLength)
                    npt.assert_equal(out.shape[-1], expectedNumBlocks)


    def test_spectral_features_for_1d_input(self):
        fs = 44100
        X = np.zeros(1025)

        features = pyACA.getFeatureList('spectral')
        features.remove('SpectralMfccs')
        features.remove('SpectralPitchChroma')

        for feature in features:
            with self.subTest(msg=feature):
                featureFunc = getattr(pyACA, "Feature" + feature)
                out = featureFunc(X, fs)
                npt.assert_equal(out.shape, ())

        with self.subTest(msg='SpectralMfccs'):
            out = pyACA.FeatureSpectralMfccs(X, fs)
            npt.assert_equal(out.shape, (13,))

        with self.subTest(msg='SpectralPitchChroma'):
            out = pyACA.FeatureSpectralPitchChroma(X, fs)
            npt.assert_equal(out.shape, (12,))


    def test_spectral_features_for_2d_input(self):
        fs = 44100
        features = pyACA.getFeatureList('spectral')
        features.remove('SpectralMfccs')
        features.remove('SpectralPitchChroma')

        # Test for 2d input with only one block
        X = np.zeros((1025, 1))

        for feature in features:
            with self.subTest(msg=feature):
                featureFunc = getattr(pyACA, "Feature" + feature)
                out = featureFunc(X, fs)
                npt.assert_equal(out.shape, (1,))

        with self.subTest(msg='SpectralMfccs'):
            out = pyACA.FeatureSpectralMfccs(X, fs)
            npt.assert_equal(out.shape, (13, 1))

        with self.subTest(msg='SpectralPitchChroma'):
            out = pyACA.FeatureSpectralPitchChroma(X, fs)
            npt.assert_equal(out.shape, (12, 1))

        # Test for 2d input with only multiple block
        X = np.zeros((1025, 16))

        for feature in features:
            with self.subTest(msg=feature):
                featureFunc = getattr(pyACA, "Feature" + feature)
                out = featureFunc(X, fs)
                npt.assert_equal(out.shape, (16,))

        with self.subTest(msg='SpectralMfccs'):
            out = pyACA.FeatureSpectralMfccs(X, fs)
            npt.assert_equal(out.shape, (13, 16))

        with self.subTest(msg='SpectralPitchChroma'):
            out = pyACA.FeatureSpectralPitchChroma(X, fs)
            npt.assert_equal(out.shape, (12, 16))


    def calcNumBlocks(self, inputSize, blockLen, hopLen):
        return np.floor((inputSize - blockLen) / hopLen + 1)



# -*- coding: utf-8 -*-
"""
computes the spectral rolloff from the magnitude spectrum
  Args:
    X: spectrogram (dimension FFTLength X Observations)
    f_s: sample rate of audio data
    kappa: cutoff ratio

  Returns:
    vsr spectral rolloff (in Hz)
"""

import numpy as np


def FeatureSpectralRolloff(X, f_s, kappa=0.85):

    norm = X.sum(axis=0, keepdims=True)
    norm[norm == 0] = 1

    X = np.cumsum(X, axis=0) / norm

    vsr = np.argmax(X >= kappa, axis=0)

    # convert from index to Hz
    vsr = vsr / (X.shape[0] - 1) * f_s / 2

    return vsr



# -*- coding: utf-8 -*-
"""
computes the lag of the autocorrelation function

  Args:
    x: audio signal
    iBlockLength: block length in samples
    iHopLength: hop length in samples
    f_s: sample rate of audio data (unused)

  Returns:
      f frequency
      t time stamp for the frequency value
"""

import numpy as np
import math


def PitchTimeAcf(x, iBlockLength, iHopLength, f_s):

    # initialize
    f_max = 2000
    fMinThresh = .35
    iNumOfBlocks = math.ceil(x.size / iHopLength)

    # compute time stamps
    t = (np.arange(0, iNumOfBlocks) * iHopLength + (iBlockLength / 2)) / f_s

    # allocate memory
    f = np.zeros(iNumOfBlocks)

    for n in range(0, iNumOfBlocks):
        eta_min = int(round(f_s / f_max)) - 1

        i_start = n * iHopLength
        i_stop = np.min([x.size - 1, i_start + iBlockLength - 1])

        # calculate the acf
        if not x[np.arange(i_start, i_stop + 1)].sum():
            continue
        else:
            x_tmp = x[np.arange(i_start, i_stop + 1)]
            afCorr = np.correlate(x_tmp, x_tmp, "full") / np.dot(x_tmp, x_tmp)

        afCorr = afCorr[np.arange(iBlockLength, afCorr.size)]

        # update eta_min to avoid main lobe
        eta_tmp = np.argmax(afCorr < fMinThresh)
        eta_min = np.max([eta_min, eta_tmp])

        afDeltaCorr = np.diff(afCorr)
        eta_tmp = np.argmax(afDeltaCorr > 0)
        eta_min = np.max([eta_min, eta_tmp])

        # find the coefficients specified in eta
        f[n] = np.argmax(afCorr[np.arange(eta_min + 1, afCorr.size)]) + 1

        # convert to Hz
        f[n] = f_s / (f[n] + eta_min + 1)

    return (f, t)



# -*- coding: utf-8 -*-
"""
helper function: compute periodic von-Hann window

  from https://www.programcreek.com/python/example/15589/numpy.hanning

  Args:
    window_length: The number of points in the returned window.

  Returns:
    A 1D np.array containing the periodic hann window.
"""

import numpy as np


def ToolComputeHann(iWindowLength):
    return 0.5 - (0.5 * np.cos(2 * np.pi / iWindowLength * np.arange(iWindowLength)))



# -*- coding: utf-8 -*-
"""
computes the maximum of the Harmonic Product Spectrum

  Args:
    X: spectrogram (dimension FFTLength X Observations)
    f_s: sample rate of audio data

  Returns:
    f HPS maximum location (in Hz)
"""

import numpy as np


def PitchSpectralHps(X, f_s):

    # initialize
    iOrder = 4
    f_min = 300
    f = np.zeros(X.shape[1])

    iLen = int((X.shape[0] - 1) / iOrder)
    afHps = X[np.arange(0, iLen), :]
    k_min = int(round(f_min / f_s * 2 * (X.shape[0] - 1)))

    # compute the HPS
    for j in range(1, iOrder):
        X_d = X[::(j + 1), :]
        afHps *= X_d[np.arange(0, iLen), :]

    f = np.argmax(afHps[np.arange(k_min, afHps.shape[0])], axis=0)

    # find max index and convert to Hz
    f = (f + k_min) / (X.shape[0] - 1) * f_s / 2

    return (f)



# -*- coding: utf-8 -*-
"""
computeKey

computes the musical key of an input audio file
  Args:
      afAudioData: array with floating point audio data.
      f_s: sample rate
      afWindow: FFT window of length iBlockLength (default: hann)
      iBlockLength: internal block length (default: 4096 samples)
      iHopLength: internal hop length (default: 2048 samples)

  Returns:
      key string
"""

import numpy as np
from scipy.signal import spectrogram

from pyACA.ToolComputeHann import ToolComputeHann
from pyACA.FeatureSpectralPitchChroma import FeatureSpectralPitchChroma
from pyACA.ToolPreprocAudio import ToolPreprocAudio
from pyACA.ToolReadAudio import ToolReadAudio

def computeKey(afAudioData, f_s, afWindow=None, iBlockLength=4096, iHopLength=2048):

    # compute window function for FFT
    if afWindow is None:
        afWindow = ToolComputeHann(iBlockLength)

    assert(afWindow.shape[0] == iBlockLength), "parameter error: invalid window dimension"

    # key names
    cKeyNames = np.array(['C Maj', 'C# Maj', 'D Maj', 'D# Maj', 'E Maj', 'F Maj', 'F# Maj', 'G Maj', 'G# Maj', 'A Maj', 'A# Maj', 'B Maj',
                         'c min', 'c# min', 'd min', 'd# min', 'e min', 'f min', 'f# min', 'g min', 'g# min', 'a min', 'a# min', 'b min'])

    # template pitch chroma (Krumhansl major/minor), normalized to a sum of 1
    t_pc = np.array([[6.35, 2.23, 3.48, 2.33, 4.38, 4.09, 2.52, 5.19, 2.39, 3.66, 2.29, 2.88],
                    [6.33, 2.68, 3.52, 5.38, 2.60, 3.53, 2.54, 4.75, 3.98, 2.69, 3.34, 3.17]])
    t_pc = t_pc / t_pc.sum(axis=1, keepdims=True)

    # pre-processing
    afAudioData = ToolPreprocAudio(afAudioData, iBlockLength)

    # in the real world, we would do this block by block...
    [f, t, X] = spectrogram(afAudioData,
                            f_s,
                            afWindow,
                            iBlockLength,
                            iBlockLength - iHopLength,
                            iBlockLength,
                            False,
                            True,
                            'spectrum')

    #  scale the same as for matlab
    X = np.sqrt(X / 2)

    # compute instantaneous pitch chroma
    v_pc = FeatureSpectralPitchChroma(X, f_s)

    # average pitch chroma
    v_pc = v_pc.mean(axis=1)
    # compute manhattan distances for modes (major and minor)
    d = np.zeros(t_pc.shape)
    v_pc = np.concatenate((v_pc, v_pc), axis=0).reshape(2, 12)
    for i in range(0, 12):
        d[:, i] = np.sum(np.abs(v_pc - np.roll(t_pc, i, axis=1)), axis=1)

    # get unwrapped key index
    iKeyIdx = d.argmin()

    cKey = cKeyNames[iKeyIdx]

    return (cKey)


def computeKeyCl(cPath):
    
    [f_s, afAudioData] = ToolReadAudio(cPath)
    # afAudioData = np.sin(2*np.pi * np.arange(f_s*1)*440./f_s)

    cKey = computeKey(afAudioData, f_s)
    print("\ndetected key: ", cKey)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description='Compute key of wav file')
    parser.add_argument('--infile', metavar='path', required=False,
                        help='path to input audio file')

    args = parser.parse_args()
    cPath = args.infile

    # only for debugging
    if not cPath:
        cPath = "c:/temp/test.wav"

    # call the function
    computeKeyCl(cPath)



# -*- coding: utf-8 -*-
"""
computes the lag of the amdf function

  Args:
    x: audio signal
    iBlockLength: block length in samples
    iHopLength: hop length in samples
    f_s: sample rate of audio data (unused)

  Returns:
      f frequency
      t time stamp for the frequency value
"""

import numpy as np
import math


def PitchTimeAmdf(x, iBlockLength, iHopLength, f_s):
    # initialize
    f_max = 2000
    f_min = 50
    iNumOfBlocks = math.ceil(x.size / iHopLength)

    # compute time stamps
    t = (np.arange(0, iNumOfBlocks) * iHopLength + (iBlockLength / 2)) / f_s

    # allocate memory
    f = np.zeros(iNumOfBlocks)

    eta_min = int(round(f_s / f_max)) - 1
    eta_max = int(round(f_s / f_min)) - 1

    for n in range(0, iNumOfBlocks):

        i_start = n * iHopLength
        i_stop = np.min([x.size - 1, i_start + iBlockLength - 1])

        # calculate the acf
        if not x[np.arange(i_start, i_stop + 1)].sum():
            continue
        else:
            x_tmp = x[np.arange(i_start, i_stop + 1)]
            afCorr = computeAmdf(x_tmp, eta_max)

        # find the coefficients specified in eta
        f[n] = np.argmin(afCorr[np.arange(eta_min + 1, afCorr.size)]) + 1

        # convert to Hz
        f[n] = f_s / (f[n] + eta_min + 1)

    return (f, t)


def computeAmdf(x, eta_max):
    K = x.shape[0]

    if K <= 0:
        return 0

    afAmdf = np.ones(K)

    for eta in range(0, np.min([K, eta_max + 1])):
        afAmdf[eta] = np.sum(np.abs(x[np.arange(0, K - 1 - eta)] - x[np.arange(eta + 1, K)])) / K

    return (afAmdf)



# -*- coding: utf-8 -*-
"""
computes the pitch chroma from the magnitude spectrum

  Args:
    X: spectrogram (dimension FFTLength X Observations)
    f_s: sample rate of audio data

  Returns:
    v_pc pitch chroma
"""

import numpy as np
import math


def FeatureSpectralPitchChroma(X, f_s):

    isSpectrum = X.ndim == 1
    if isSpectrum:
        X = np.expand_dims(X, axis=1)

    # allocate memory
    v_pc = np.zeros([12, X.shape[1]])

    # generate filter matrix
    H = generatePcFilters(X.shape[0], f_s)

    # compute pitch chroma
    v_pc = np.dot(H, X**2)

    # norm pitch chroma to a sum of 1 but avoid div by zero
    norm = v_pc.sum(axis=0, keepdims=True)
    norm[norm == 0] = 1
    v_pc = v_pc / norm

    return np.squeeze(v_pc) if isSpectrum else v_pc


def generatePcFilters(iSpecLength, f_s):

    # initialization at C4
    f_mid = 261.63
    iNumOctaves = 4
    iNumPitchesPerOctave = 12

    # sanity check
    while (f_mid * 2**iNumOctaves > f_s / 2.):
        iNumOctaves = iNumOctaves - 1

    H = np.zeros([iNumPitchesPerOctave, iSpecLength])

    # for each pitch class i create weighting factors in each octave j
    for i in range(0, iNumPitchesPerOctave):
        afBounds = np.array([2**(-1 / (2 * iNumPitchesPerOctave)), 2**(1 / (2 * iNumPitchesPerOctave))]) * f_mid * 2 * (iSpecLength - 1) / f_s
        for j in range(0, iNumOctaves):
            iBounds = np.array([math.ceil(2**j * afBounds[0]), math.ceil(2**j * afBounds[1])])
            H[i, range(iBounds[0], iBounds[1])] = 1 / (iBounds[1] - iBounds[0])

        # increment to next semi-tone
        f_mid = f_mid * 2**(1 / iNumPitchesPerOctave)

    return (H)



# -*- coding: utf-8 -*-
"""
computes the spectral decrease from the magnitude spectrum

  Args:
    X: spectrogram (dimension FFTLength X Observations)
    f_s: sample rate of audio data

  Returns:
    vsk spectral decrease
"""

import numpy as np


def FeatureSpectralDecrease(X, f_s):

    # compute index vector
    kinv = np.arange(0, X.shape[0])
    kinv[0] = 1
    kinv = 1 / kinv

    norm = X[1:].sum(axis=0, keepdims=True)
    norm[norm == 0] = 1

    # compute slope
    vsc = np.dot(kinv, X - X[0]) / norm

    return np.squeeze(vsc, axis=0)



# -*- coding: utf-8 -*-
"""
helper function: convert Hz to Mel scale

  Args:
    fInHz: The frequency to be converted, can be scalar or vector
    cModel: The name of the model ('Fant' [default], 'Shaughnessy', 'Umesh')

  Returns:
    Mel values of the input dimension 
"""

import numpy as np
import math


def ToolFreq2Mel(fInHz, cModel = 'Fant'):
    def acaFant_scalar(f):
        return 1000 * math.log2(1 + f/1000)
    def acaShaughnessy_scalar(f):
        return 2595 * math.log10(1 + f/700)
    def acaUmesh_scalar(f):
        return f/(2.4e-4*f + 0.741)

    f = np.asarray(fInHz)
    if f.ndim == 0:
        if cModel == 'Shaughnessy':
            return acaShaughnessy_scalar(f)
        elif cModel == 'Umesh':
            return acaUmesh_scalar(f)
        else:
            return acaFant_scalar(f)

    fMel = np.zeros(f.shape)
    if cModel == 'Shaughnessy':
        for k,fi in enumerate(f):
            fMel[k] =  acaShaughnessy_scalar(fi)
    elif cModel == 'Umesh':
        for k,fi in enumerate(f):
            fMel[k] =  acaUmesh_scalar(fi)
    else:
        for k,fi in enumerate(f):
            fMel[k] =  acaFant_scalar(fi)
            
    return (fMel)



"""
helper function: Creates blocks from given array. This method creates a block only if there is enough input data to
fill the block. This means if there isn't enough data to fill the last block then that lst chunk of input data will
be discarded. To avoid losing data, you should pad the input with zeros of at least iBlockLength length.

  Args:
    afAudioData: 1D np.array
    iBlockLength: block length
    iHopLength: hop length

  Returns:
    A 2D np.array containing the blocked data of shape (iNumOfBlocks, iBlockLength).
"""

import numpy as np


def ToolBlockAudio(afAudioData, iBlockLength, iHopLength):

    iNumOfBlocks = np.floor((afAudioData.shape[0] - iBlockLength) / iHopLength + 1).astype(int)

    if iNumOfBlocks < 1:
        return np.array([])
    return np.vstack([np.array(afAudioData[i*iHopLength:i*iHopLength+iBlockLength]) for i in range(iNumOfBlocks)])






# -*- coding: utf-8 -*-
"""
computes the RMS of a time domain signal

  Args:
    x: audio signal
    iBlockLength: block length in samples
    iHopLength: hop length in samples
    f_s: sample rate of audio data (unused)

  Returns:
    vrms autocorrelation maximum
    t time stamp

"""

import numpy as np
import pyACA


def FeatureTimeRms(x, iBlockLength, iHopLength, f_s):

    # create blocks
    xBlocks = pyACA.ToolBlockAudio(x, iBlockLength, iHopLength)

    # number of results
    iNumOfBlocks = xBlocks.shape[0]

    # compute time stamps
    t = (np.arange(0, iNumOfBlocks) * iHopLength + (iBlockLength / 2)) / f_s

    # allocate memory
    vrms = np.zeros(iNumOfBlocks)

    for n, block in enumerate(xBlocks):
        # calculate the rms
        vrms[n] = np.sqrt(np.dot(block, block) / block.size)

    # convert to dB
    epsilon = 1e-5  # -100dB

    vrms[vrms < epsilon] = epsilon
    vrms = 20 * np.log10(vrms)

    return vrms, t



# -*- coding: utf-8 -*-
"""
computeFeature

computes a feature from the audio data
supported features are:
    'SpectralCentroid',
    'SpectralCrestFactor',
    'SpectralDecrease',
    'SpectralFlatness',
    'SpectralFlux',
    'SpectralKurtosis',
    'SpectralMfccs',
    'SpectralPitchChroma',
    'SpectralRolloff',
    'SpectralSkewness',
    'SpectralSlope',
    'SpectralSpread',
    'SpectralTonalPowerRatio',
    'TimeAcfCoeff',
    'TimeMaxAcf',
    'TimePeakEnvelope',
    'TimePredictivityRatio',
    'TimeRms',
    'TimeStd',
    'TimeZeroCrossingRate',
  Args:
      cFeatureName: feature to compute, e.g. 'SpectralSkewness'
      afAudioData: array with floating point audio data.
      f_s: sample rate
      afWindow: FFT window of length iBlockLength (default: hann)
      iBlockLength: internal block length (default: 4096 samples)
      iHopLength: internal hop length (default: 2048 samples)

  Returns:
      feature value v
      time stamp t
"""

import numpy as np
from scipy.signal import spectrogram
import matplotlib.pyplot as plt

import pyACA
from pyACA.ToolPreprocAudio import ToolPreprocAudio
from pyACA.ToolComputeHann import ToolComputeHann
from pyACA.ToolReadAudio import ToolReadAudio


def computeFeature(cFeatureName, afAudioData, f_s, afWindow=None, iBlockLength=4096, iHopLength=2048):
 
    #mypackage = __import__(".Feature" + cFeatureName, package="pyACA")
    hFeatureFunc = getattr(pyACA, "Feature" + cFeatureName)

    # pre-processing
    afAudioData = ToolPreprocAudio(afAudioData, iBlockLength)

    if isSpectral(cFeatureName):
        # compute window function for FFT
        if afWindow is None:
            afWindow = ToolComputeHann(iBlockLength)

        assert(afWindow.shape[0] == iBlockLength), "parameter error: invalid window dimension"

        # in the real world, we would do this block by block...
        [f, t, X] = spectrogram(afAudioData,
                                fs=f_s,
                                window=afWindow,
                                nperseg=iBlockLength,
                                noverlap=iBlockLength - iHopLength,
                                nfft=iBlockLength,
                                detrend=False,
                                return_onesided=True,
                                scaling='spectrum')

        # we just want the magnitude spectrum...
        X = np.sqrt(X / 2)

        # compute instantaneous feature
        v = hFeatureFunc(X, f_s)

    if isTemporal(cFeatureName):
        [v, t] = hFeatureFunc(afAudioData, iBlockLength, iHopLength, f_s)
        # [v, t] = hFeatureFunc(afAudioData, iBlockLength, iHopLength, f_s, np.array([2, 3]))

    return (v, t)


#######################################################
# helper functions
def isSpectral(cName):
    bResult = False
    if "Spectral" in cName:
        bResult = True

    return (bResult)


def isTemporal(cName):
    bResult = False
    if "Time" in cName:
        bResult = True

    return (bResult)


def computeFeatureCl(cPath, cFeatureName, bPlotOutput = False):

    # read audio file
    [f_s, afAudioData] = ToolReadAudio(cPath)
    
    # for debugging
    #afAudioData = np.sin(2*np.pi * np.arange(f_s*1)*440./f_s)

    # compute feature
    [v, t] = computeFeature(cFeatureName, afAudioData, f_s, None, 1024, 256)

    # plot feature output
    if bPlotOutput:
        plt.plot(t, v)

    return (v, t)


if __name__ == "__main__":
    import argparse

    # add command line args and parse them
    parser = argparse.ArgumentParser(description='Compute key of wav file')
    parser.add_argument('--infile', metavar='path', required=False,
                        help='path to input audio file')
    parser.add_argument('--featurename', metavar='string', required=False,
                        help='feature name in the format SpectralPitchChroma')
    parser.add_argument('--plotoutput', metavar='bool', required=False,
                        help='option to plot the output')

    # retrieve command line args
    args = parser.parse_args()
    cPath = args.infile
    cFeatureName = args.featurename
    bPlotOutput = args.plotoutput

    # only for debugging
    if __debug__:
        if not cPath:
            cPath = "c:/temp/test.wav"
        if not cFeatureName:
            cFeatureName = "SpectralCentroid"
        if not bPlotOutput:
            bPlotOutput = False

    # call the function
    computeFeatureCl(cPath, cFeatureName, bPlotOutput)




import numpy as np


# see function mfcc.m from Slaneys Auditory Toolbox
def ToolMfccFb(iFftLength, f_s):

    # initialization
    f_start = 133.3333

    iNumLinFilters = 13
    iNumLogFilters = 27
    iNumFilters = iNumLinFilters + iNumLogFilters

    linearSpacing = 66.66666666
    logSpacing = 1.0711703

    # compute band frequencies
    f = np.zeros(iNumFilters + 2)
    f[np.arange(0, iNumLinFilters)] = f_start + np.arange(0, iNumLinFilters) * linearSpacing
    f[np.arange(iNumLinFilters, iNumFilters + 2)] = f[iNumLinFilters - 1] * logSpacing**np.arange(1, iNumLogFilters + 3)

    # sanity check
    if f[iNumLinFilters - 1] >= f_s / 2:
        f = f[f < f_s / 2]
        iNumFilters = f.shape[0] - 2

    f_l = f[np.arange(0, iNumFilters)]
    f_c = f[np.arange(1, iNumFilters + 1)]
    f_u = f[np.arange(2, iNumFilters + 2)]

    # allocate memory for filters and set max amplitude
    H = np.zeros([iNumFilters, iFftLength])
    afFilterMax = 2 / (f_u - f_l)
    f_k = np.arange(0, iFftLength) / (iFftLength - 1) * f_s / 2

    # compute the transfer functions
    for c in range(0, iNumFilters):
        # lower filter slope
        i_l = np.argmax(f_k > f_l[c])
        i_u = np.max([0, np.argmin(f_k <= f_c[c]) - 1])
        H[c, np.arange(i_l, i_u + 1)] = afFilterMax[c] * (f_k[np.arange(i_l, i_u + 1)] - f_l[c]) / (f_c[c] - f_l[c])
        # upper filter slope
        i_l = i_u + 1
        i_u = np.max([0, np.argmin(f_k < f_u[c]) - 1])
        H[c, np.arange(i_l, i_u + 1)] = afFilterMax[c] * (f_u[c] - f_k[np.arange(i_l, i_u + 1)]) / (f_u[c] - f_c[c])

    return (H)



# -*- coding: utf-8 -*-
"""
computes f_0 through zero crossing distances

  Args:
    x: audio signal
    iBlockLength: block length in samples
    iHopLength: hop length in samples
    f_s: sample rate of audio data (unused)

  Returns:
      f frequency
      t time stamp for the frequency value
"""

import numpy as np
import math


def PitchTimeZeroCrossings(x, iBlockLength, iHopLength, f_s):

    # initialize
    iNumOfBlocks = math.ceil(x.size / iHopLength)
    f = np.zeros(iNumOfBlocks)

    # compute time stamps
    t = (np.arange(0, iNumOfBlocks) * iHopLength + (iBlockLength / 2)) / f_s

    for n in range(0, iNumOfBlocks):

        i_start = n * iHopLength
        i_stop = np.min([x.size - 1, i_start + iBlockLength - 1])

        # get current block
        if not x[np.arange(i_start, i_stop + 1)].sum():
            continue
        else:
            x_tmp = x[np.arange(i_start, i_stop + 1)]

        # compute zero crossing indices
        x_tmp = x_tmp[np.arange(0, iBlockLength - 1)] * x_tmp[np.arange(1, iBlockLength)]
        i_tmp = np.diff(np.argwhere(x_tmp < 0), axis=0)

        # average distance of zero crossings indicates half period
        if i_tmp.size:
            f[n] = f_s / np.mean(2 * i_tmp)

    return (f, t)



#
name = "pyACA"
from .computeBeatHisto import computeBeatHisto
from .computeFeature import computeFeature
from .computeKey import computeKey
from .computeNoveltyFunction import computeNoveltyFunction
from .computePitch import computePitch
from .computeBeatHisto import computeBeatHistoCl
from .computeFeature import computeFeatureCl
from .computeKey import computeKeyCl
from .computeNoveltyFunction import computeNoveltyFunctionCl
from .computePitch import computePitchCl
from .ToolSimpleDtw import ToolSimpleDtw
from .ToolReadAudio import ToolReadAudio
from .ToolPreprocAudio import ToolPreprocAudio
from .ToolBlockAudio import ToolBlockAudio
from .FeatureSpectralCentroid import FeatureSpectralCentroid
from .FeatureSpectralCrestFactor import FeatureSpectralCrestFactor
from .FeatureSpectralDecrease import FeatureSpectralDecrease
from .FeatureSpectralFlatness import FeatureSpectralFlatness
from .FeatureSpectralFlux import FeatureSpectralFlux
from .FeatureSpectralKurtosis import FeatureSpectralKurtosis
from .FeatureSpectralMfccs import FeatureSpectralMfccs
from .FeatureSpectralPitchChroma import FeatureSpectralPitchChroma
from .FeatureSpectralRolloff import FeatureSpectralRolloff
from .FeatureSpectralSkewness import FeatureSpectralSkewness
from .FeatureSpectralSlope import FeatureSpectralSlope
from .FeatureSpectralSpread import FeatureSpectralSpread
from .FeatureSpectralTonalPowerRatio import FeatureSpectralTonalPowerRatio
from .FeatureTimeAcfCoeff import FeatureTimeAcfCoeff
from .FeatureTimeMaxAcf import FeatureTimeMaxAcf
from .FeatureTimePeakEnvelope import FeatureTimePeakEnvelope
from .FeatureTimeRms import FeatureTimeRms
from .FeatureTimeStd import FeatureTimeStd
from .FeatureTimeZeroCrossingRate import FeatureTimeZeroCrossingRate
from .NoveltyFlux import NoveltyFlux
from .NoveltyHainsworth import NoveltyHainsworth
from .NoveltyLaroche import NoveltyLaroche
from .PitchSpectralAcf import PitchSpectralAcf
from .PitchSpectralHps import PitchSpectralHps
from .PitchTimeAcf import PitchTimeAcf
from .PitchTimeAmdf import PitchTimeAmdf
from .PitchTimeAuditory import PitchTimeAuditory
from .PitchTimeZeroCrossings import PitchTimeZeroCrossings
from .getFeatureList import getFeatureList

features = [
    "SpectralCentroid",
    "SpectralCrestFactor",
    "SpectralDecrease",
    "SpectralFlatness",
    "SpectralFlux",
    "SpectralKurtosis",
    "SpectralMfccs",
    "SpectralPitchChroma",
    "SpectralRolloff",
    #"SpectralSkewness",
    "SpectralSlope",
    "SpectralTonalPowerRatio",
    #"TimeAcfCoeff",
    #"TimeMaxAcf",
    #"TimePeakEnvelope",
    #"TimeRms",
    #"TimeStd",
    #"TimeZeroCrossingRate"
]


# -*- coding: utf-8 -*-
"""
computes the spectral flatness from the magnitude spectrum

  Args:
    X: spectrogram (dimension FFTLength X Observations)
    f_s: sample rate of audio data

  Returns:
    vtf spectral flatness
"""

import numpy as np


def FeatureSpectralFlatness(X, f_s):

    norm = X.mean(axis=0, keepdims=True)
    norm[norm == 0] = 1

    X = np.log(X + 1e-20)

    vtf = np.exp(X.mean(axis=0, keepdims=True)) / norm

    return np.squeeze(vtf, axis=0)



# -*- coding: utf-8 -*-
"""
NoveltyFlux

computes the novelty measure per Spectral Flux

  Args:
      X: spectrogram (dimension FFTLength X Observations)
      afAudioData: array with floating point audio data.
      f_s: sample rate

  Returns:
      d_flux novelty measure

"""

from .FeatureSpectralFlux import FeatureSpectralFlux


def NoveltyFlux(X, f_s):
    d_flux = FeatureSpectralFlux(X, f_s)

    return (d_flux)



# -*- coding: utf-8 -*-
"""
helper function: read audio from wav

  Args:
    cAudioFilePath: path to audio file

  Returns:
    sample rate (scalar)
    audio (array)
"""

from scipy.io.wavfile import read as wavread


def ToolReadAudio(cAudioFilePath):
    [samplerate, x] = wavread(cAudioFilePath)

    if x.dtype == 'float32':
        audio = x
    else:
        # change range to [-1,1)
        if x.dtype == 'uint8':
            nbits = 8
        elif x.dtype == 'int16':
            nbits = 16
        elif x.dtype == 'int32':
            nbits = 32

        audio = x / float(2**(nbits - 1))

    # special case of unsigned format
    if x.dtype == 'uint8':
        audio = audio - 1.

    return (samplerate, audio)



# -*- coding: utf-8 -*-
"""
computes the spectral crest from the magnitude spectrum

  Args:
    X: spectrogram (dimension FFTLength X Observations)
    f_s: sample rate of audio data

  Returns:
    v spectral crest factor
"""

import numpy as np

def FeatureSpectralCrestFactor(X, f_s):

    norm = X.sum(axis=0, keepdims=True)
    norm[norm == 0] = 1

    vtsc = X.max(axis=0) / norm

    return np.squeeze(vtsc, axis=0)



# -*- coding: utf-8 -*-
"""
computes f_0 through "auditory" approach

  Args:
    x: audio signal
    iBlockLength: block length in samples
    iHopLength: hop length in samples
    f_s: sample rate of audio data (unused)

  Returns:
      f frequency
      t time stamp for the frequency value
"""

import numpy as np
import math
from scipy.signal import filtfilt
from scipy.signal import find_peaks

from .ToolGammatoneFb import ToolGammatoneFb


def PitchTimeAuditory(x, iBlockLength, iHopLength, f_s):

    # initialize
    iNumOfBlocks = math.ceil(x.size / iHopLength)
    f = np.zeros(iNumOfBlocks)
    f_max = 2000
    iNumBands = 20
    fLengthLpInS = 0.001

    iLengthLp = math.ceil(fLengthLpInS * f_s)

    # compute time stamps
    t = (np.arange(0, iNumOfBlocks) * iHopLength + (iBlockLength / 2)) / f_s

    # apply filterbank
    X = ToolGammatoneFb(x, f_s, iNumBands)

    # half wave rectification
    X[X < 0] = 0

    # smooth the results with a moving average filter
    b = np.ones(iLengthLp) / iLengthLp
    X = filtfilt(b, 1, X)

    for n in range(0, iNumOfBlocks):

        eta_min = int(round(f_s / f_max))
        afSumCorr = np.zeros(iBlockLength - 1)
        x_tmp = np.zeros(iBlockLength)

        i_start = n * iHopLength
        i_stop = np.min([x.size - 1, i_start + iBlockLength - 1])

        # compute ACF per band and summarize
        for k in range(0, iNumBands):
            # get current block
            if X[k, np.arange(i_start, i_stop + 1)].sum() < 1e-20:
                continue
            else:
                x_tmp[np.arange(0, i_stop - i_start + 1)] = X[k, np.arange(i_start, i_stop + 1)]

            afCorr = np.correlate(x_tmp, x_tmp, "full") / np.dot(x_tmp, x_tmp)

            # aggregate bands with simple sum before peak picking
            afSumCorr += afCorr[np.arange(iBlockLength, afCorr.size)]

        if afSumCorr.sum() < 1e-20:
            continue

        # find the highest local maximum
        iPeaks = find_peaks(afSumCorr, height=0)
        if iPeaks[0].size:
            eta_min = np.max([eta_min, iPeaks[0][0] - 1])
        f[n] = np.argmax(afSumCorr[np.arange(eta_min + 1, afSumCorr.size)]) + 1

        # convert to Hz
        f[n] = f_s / (f[n] + eta_min + 1)

    return (f, t)



# -*- coding: utf-8 -*-
"""
helper function: compute periodic von-Hann window

  Args:
    fInHz: The frequency to be converted, can be scalar or vector
    fA4InHz: The reference tuning frequency (default: 440Hz)

  Returns:
    Midi values of the input dimension (floating point)
"""

import numpy as np


def ToolMidi2Freq(pInMidi, fA4InHz = 440):
    def convert_midi2freq_scalar(p, fA4InHz):
 
        if f <= 0:
            return 0
        else:
            return fA4InHz * 2**((p-69) / 12)

    pInMidi = np.asarray(pInMidi)
    if pInMidi.ndim == 0:
       return convert_midi2freq_scalar(pInMidi,fA4InHz)

    fInHz = np.zeros(pInMidi.shape)
    for k,p in enumerate(pInMidi):
        fInHz[k] =  convert_midi2freq_scalar(p,fA4InHz)
            
    return (midi)



# -*- coding: utf-8 -*-
"""
computes the zero crossing rate of a time domain signal

  Args:
    x: audio signal
    iBlockLength: block length in samples
    iHopLength: hop length in samples
    f_s: sample rate of audio data (unused)

  Returns:
    vzc zero crossing rate
    t time stamp

"""

import numpy as np
import pyACA


def FeatureTimeZeroCrossingRate(x, iBlockLength, iHopLength, f_s):

    # create blocks
    xBlocks = pyACA.ToolBlockAudio(x, iBlockLength, iHopLength)

    # number of results
    iNumOfBlocks = xBlocks.shape[0]

    # compute time stamps
    t = (np.arange(0, iNumOfBlocks) * iHopLength + (iBlockLength / 2)) / f_s

    # allocate memory
    vzc = np.zeros(iNumOfBlocks)

    for n, block in enumerate(xBlocks):
        # calculate the zero crossing rate
        vzc[n] = 0.5 * np.mean(np.abs(np.diff(np.sign(block))))

    return vzc, t



# -*- coding: utf-8 -*-
"""
NoveltyHainsworth

computes the novelty measure used by Hainsworth

  Args:
      X: spectrogram (dimension FFTLength X Observations)
      f_s: sample rate

  Returns:
      d_hai novelty measure

"""

import numpy as np


def NoveltyHainsworth(X, f_s):

    epsilon = 1e-5

    # difference spectrum (set first diff to zero)
    X = np.c_[X[:, 0], np.sqrt(X)]
    X[X <= 0] = epsilon

    afDiff = np.log2(X[:, np.arange(1, X.shape[1])] / X[:, np.arange(0, X.shape[1] - 1)])

    # flux
    d_hai = np.sum(afDiff, axis=0) / X.shape[0]

    return (d_hai)



# -*- coding: utf-8 -*-
"""
helper function: convert Hz to Bark scale

  Args:
    fInHz: The frequency to be converted, can be scalar or vector
    cModel: The name of the model ('Schroeder' [default], 'Terhardt', 'Zwicker', 'Traunmuller')

  Returns:
    Bark values of the input dimension 
"""

import numpy as np
import math


def ToolFreq2Bark(fInHz, cModel = 'Schroeder'):
    def acaSchroeder_scalar(f):
        return 7 * math.asinh(f/650)
    def acaTerhardt_scalar(f):
        return 13.3 * math.atan(0.75 * f/1000)
    def acaZwicker_scalar(f):
        return 13 * math.atan(0.76 * f/1000) + 3.5 * math.atan(f/7500)
    def acaTraunmuller_scalar(f):
        return 26.81/(1+1960./f) - 0.53

    f = np.asarray(fInHz)
    if f.ndim == 0:
        if cModel == 'Terhardt':
            return acaTerhardt_scalar(f)
        elif cModel == 'Zwicker':
            return acaZwicker_scalar(f)
        elif cModel == 'Traunmuller':
            return acaTraunmuller_scalar(f)
        else:
            return acaSchroeder_scalar(f)

    fBark = np.zeros(f.shape)
    if cModel == 'Terhardt':
        for k,fi in enumerate(f):
            fBark[k] =  acaTerhardt_scalar(fi)
    elif cModel == 'Zwicker':
        for k,fi in enumerate(f):
            fBark[k] =  acaZwicker_scalar(fi)
    elif cModel == 'Traunmuller':
        for k,fi in enumerate(f):
            fBark[k] =  acaTraunmuller_scalar(fi)
    else:
        for k,fi in enumerate(f):
            fBark[k] =  acaSchroeder_scalar(fi)
            
    return (fBark)



# -*- coding: utf-8 -*-
"""
helper function: computes path through a distance matrix with simple Dynamic Time Warping

  Args:
    D: non-negative distance matrix

  Returns:
    p: path with matrix indices
    C: cost matrix
"""
import numpy as np


def ToolSimpleDtw(D):

    # init directions for back-tracking
    iDec = np.array([[-1, -1], [-1, 0], [0, -1]])  # DeltaP contents: diag, vert, hori

    # cost initialization
    C = np.zeros(D.shape)
    C[0, :] = np.cumsum(D[0, :])
    C[:, 0] = np.cumsum(D[:, 0])

    # traceback initialization
    DeltaP = np.zeros(D.shape, dtype=int)
    DeltaP[0, :] = 2  # (0,-1)
    DeltaP[:, 0] = 1  # (-1,0)
    DeltaP[0, 0] = 0  # (-1,-1)

    # recursion
    for n_A in range(1, D.shape[0]):
        for n_B in range(1, D.shape[1]):
            # find preceding min (diag, column, row)
            DeltaP[n_A, n_B] = int(np.argmin([C[n_A - 1, n_B - 1], C[n_A - 1, n_B], C[n_A, n_B - 1]]))
            prevC_index = [n_A, n_B] + iDec[DeltaP[n_A, n_B], :]
            C[n_A, n_B] = D[n_A, n_B] + C[prevC_index[0], prevC_index[1]]

    # traceback init
    p = np.asarray(D.shape, dtype=int) - 1  # start with the last element
    n = p

    while (n[0] >= 0) or (n[1] >= 0):
        n = n + iDec[DeltaP[n[0], n[1]], :]

        # update path
        tmp = np.vstack([n, p])
        p = tmp

    return (p[1:, :], C)



# -*- coding: utf-8 -*-
"""
computes the MFCCs from the magnitude spectrum (see Slaney)

  Args:
    X: spectrogram (dimension FFTLength X Observations)
    f_s: sample rate of audio data

  Returns:
    v_mfcc mel frequency cepstral coefficients
"""

import numpy as np
from .ToolMfccFb import ToolMfccFb


def FeatureSpectralMfccs(X, f_s, iNumCoeffs=13):

    isSpectrum = X.ndim == 1
    if isSpectrum:
        X = np.expand_dims(X, axis=1)

    # allocate memory
    v_mfcc = np.zeros([iNumCoeffs, X.shape[1]])

    # generate filter matrix
    H = ToolMfccFb(X.shape[0], f_s)
    T = generateDctMatrix(H.shape[0], iNumCoeffs)

    for n in range(0, X.shape[1]):
        # compute the mel spectrum
        X_Mel = np.log10(np.dot(H, X[:, n] + 1e-20))

        # calculate the mfccs
        v_mfcc[:, n] = np.dot(T, X_Mel)

    return np.squeeze(v_mfcc) if isSpectrum else v_mfcc


# see function mfcc.m from Slaneys Auditory Toolbox
def generateDctMatrix(iNumBands, iNumCepstralCoeffs):
    T = np.cos(np.outer(np.arange(0, iNumCepstralCoeffs), (2 * np.arange(0, iNumBands) + 1)) * np.pi / 2 / iNumBands)

    T = T / np.sqrt(iNumBands / 2)
    T[0, :] = T[0, :] / np.sqrt(2)

    return (T)



import unittest
import numpy as np

import pyACA

class TestFeatures(unittest.TestCase):

    def test_spectral_centroid(self):
        X = np.zeros(1025)
        fs = 4

        # zero input
        vsc = pyACA.FeatureSpectralCentroid(X, fs)
        self.assertEqual(vsc, 0, "SC 1: Zero input incorrect")

        # one peak input
        X[512] = 1
        vsc = pyACA.FeatureSpectralCentroid(X, fs)
        self.assertEqual(vsc, 1, "SC 2: Delta input incorrect")

        # flat spec input
        X = 2*np.ones(1025)
        vsc = pyACA.FeatureSpectralCentroid(X, fs)
        self.assertEqual(vsc, 1, "SC 3: Flat input incorrect")

        # i/o dimensions
        X = np.ones([1025,4])
        vsc = pyACA.FeatureSpectralCentroid(X, fs)
        self.assertEqual(len(np.squeeze(vsc)), 4, "SC 4: output vector dimension incorrect")


    def test_spectral_crest(self):
        X = np.zeros(1025)
        fs = 4

        # zero input
        vtsc = pyACA.FeatureSpectralCrestFactor(X, fs)
        self.assertEqual(vtsc, 0, "TSC 1: Zero input incorrect")

        # one peak input
        X[512] = 1
        vtsc = pyACA.FeatureSpectralCrestFactor(X, fs)
        self.assertEqual(vtsc, 1, "TSC 2: Delta input incorrect")

        # flat spec input
        X = 2*np.ones(1025)
        vtsc = pyACA.FeatureSpectralCrestFactor(X, fs)
        self.assertEqual(vtsc, 1/len(X), "TSC 3: Flat input incorrect")

        # i/o dimensions
        X = np.ones([1025,4])
        vtsc = pyACA.FeatureSpectralCrestFactor(X, fs)
        self.assertEqual(len(np.squeeze(vtsc)), 4, "TSC 4: output vector dimension incorrect")

    def test_spectral_decrease(self):
        X = np.zeros(1025)
        fs = 4

        # zero input
        vsd = pyACA.FeatureSpectralDecrease(X, fs)
        self.assertEqual(vsd, 0, "SD 1: Zero input incorrect")

        # one peak input
        X[512] = 1
        vsd = pyACA.FeatureSpectralDecrease(X, fs)
        self.assertEqual(vsd, 1.0/512, "SD 2: Delta input incorrect")

        # flat spec input
        X = 2*np.ones(1025)
        vsd = pyACA.FeatureSpectralDecrease(X, fs)
        self.assertEqual(vsd, 0, "SD 3: Flat input incorrect")

        # i/o dimensions
        X = np.ones([1025,4])
        vsd = pyACA.FeatureSpectralDecrease(X, fs)
        self.assertEqual(len(np.squeeze(vsd)), 4, "SD 4: output vector dimension incorrect")

    def test_spectral_flatness(self):
        X = np.zeros(1025)
        fs = 4

        # zero input
        vtf = pyACA.FeatureSpectralFlatness(X, fs)
        self.assertEqual(vtf, 0, "TF 1: Zero input incorrect")

        # one peak input
        X[512] = 1
        vtf = pyACA.FeatureSpectralFlatness(X, fs)
        self.assertEqual(vtf, 0, "TF 2: Delta input")

        # flat spec input
        X = 2*np.ones(1025)
        vtf = pyACA.FeatureSpectralFlatness(X, fs)
        self.assertEqual(vtf, 1, "TF 3: Flat input incorrect")

        # i/o dimensions
        X = np.ones([1025,4])
        vtf = pyACA.FeatureSpectralFlatness(X, fs)
        self.assertEqual(len(np.squeeze(vtf)), 4, "TF 4: output vector dimension incorrect")

    #def test_spectral_flux(self):
    #    X = np.zeros(1025)
    #    fs = 4

    #    # zero input
    #    vsf = pyACA.FeatureSpectralFlatness(X, fs)
    #    self.assertEqual(vsf, 0, "TF 1: Zero input incorrect")

    #    # one peak input
    #    X[512] = 1
    #    vsf = pyACA.FeatureSpectralFlatness(X, fs)
    #    self.assertEqual(vsf, 0, "TF 2: Delta input")

    #    # flat spec input
    #    X = 2*np.ones(1025)
    #    vsf = pyACA.FeatureSpectralFlatness(X, fs)
    #    self.assertEqual(vsf, 1, "TF 3: Flat input incorrect")

    #    # i/o dimensions
    #    X = np.ones([1025,4])
    #    vsf = pyACA.FeatureSpectralFlatness(X, fs)
    #    self.assertEqual(len(np.squeeze(vsf)), 4, "TF 4: output vector dimension incorrect")


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
computes the spectral spread from the magnitude spectrum

  Args:
    X: spectrogram (dimension FFTLength X Observations)
    f_s: sample rate of audio data

  Returns:
    vss spectral spread (in Hz)
"""

import numpy as np
from .FeatureSpectralCentroid import FeatureSpectralCentroid


def FeatureSpectralSpread(X, f_s):

    isSpectrum = X.ndim == 1
    if isSpectrum:
        X = np.expand_dims(X, axis=1)

    # get spectral centroid as index
    vsc = FeatureSpectralCentroid(X, f_s) * 2 / f_s * (X.shape[0] - 1)

    # X = X**2 removed for consistency with book

    norm = X.sum(axis=0)
    norm[norm == 0] = 1

    # compute spread
    vss = np.zeros(X.shape[1])
    indices = np.arange(0, X.shape[0])
    for n in range(0, X.shape[1]):
        vss[n] = np.dot((indices - vsc[n])**2, X[:, n]) / norm[n]

    vss = np.sqrt(vss)

    # convert from index to Hz
    vss = vss / (X.shape[0] - 1) * f_s / 2

    return np.squeeze(vss) if isSpectrum else vss



# -*- coding: utf-8 -*-
"""
computeNoveltyFunction

computes the novelty function for onset detection
 supported novelty measures are:
  'Flux',
  'Laroche',
  'Hainsworth'

  Args:
      cNoveltyName: name of the novelty measure
      afAudioData: array with floating point audio data.
      f_s: sample rate
      afWindow: FFT window of length iBlockLength (default: hann)
      iBlockLength: internal block length (default: 4096 samples)
      iHopLength: internal hop length (default: 2048 samples)

  Returns:
      d novelty function
      t time stamps
      iPeaks indices of picked onset times
"""

import math
import numpy as np
import matplotlib.pyplot as plt
from scipy.signal import spectrogram
from scipy.signal import filtfilt
from scipy.signal import find_peaks

import pyACA
from pyACA.ToolPreprocAudio import ToolPreprocAudio
from pyACA.ToolComputeHann import ToolComputeHann
from pyACA.ToolReadAudio import ToolReadAudio


def computeNoveltyFunction(cNoveltyName, afAudioData, f_s, afWindow=None, iBlockLength=4096, iHopLength=512):

    # compute window function for FFT
    if afWindow is None:
        afWindow = ToolComputeHann(iBlockLength)

    assert(afWindow.shape[0] == iBlockLength), "parameter error: invalid window dimension"

    #mypackage = __import__(".Novelty" + cNoveltyName, package="pyACA")
    hNoveltyFunc = getattr(pyACA, "Novelty" + cNoveltyName)

    # initialization
    fLengthLpInS = 0.3
    iLengthLp = np.max([2, math.ceil(fLengthLpInS * f_s / iHopLength)])

    # pre-processing
    afAudioData = ToolPreprocAudio(afAudioData, iBlockLength)

    # in the real world, we would do this block by block...
    [f, t, X] = spectrogram(afAudioData,
                            f_s,
                            afWindow,
                            iBlockLength,
                            iBlockLength - iHopLength,
                            iBlockLength,
                            False,
                            True,
                            'spectrum')

    #  scale the same as for matlab
    X = np.sqrt(X / 2)

    # novelty function
    d = hNoveltyFunc(X, f_s)

    # smooth novelty function
    b = np.ones(10) / 10
    d = filtfilt(b, 1, d)
    d[d < 0] = 0

    # compute threshold
    b = np.ones(iLengthLp) / iLengthLp
    G_T = .5 * np.mean(d[np.arange(1, d.shape[0])]) + filtfilt(b, 1, d)

    # find local maxima above the threshold
    iPeaks = find_peaks(d - G_T, height=0)

    return (d, t, iPeaks[0])


def computeNoveltyFunctionCl(cPath, cNoveltyName):
    
    [f_s, afAudioData] = ToolReadAudio(cPath)
    # afAudioData = np.sin(2*np.pi * np.arange(f_s*1)*440./f_s)
    [d, t, iPeaks] = computeNoveltyFunction(cNoveltyName, afAudioData, f_s)
    # plot feature output
    if bPlotOutput:
        plt.plot(t, d)
    return (d, t, iPeaks)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description='Compute key of wav file')
    parser.add_argument('--infile', metavar='path', required=False,
                        help='path to input audio file')
    parser.add_argument('--noveltyname', metavar='string', required=False,
                        help='novelty measure name in the format NoveltyFlux')
    parser.add_argument('--plotoutput', metavar='bool', required=False,
                        help='option to plot the output')

    # retrieve command line args
    args = parser.parse_args()
    cPath = args.infile
    cNoveltyName = args.noveltyname
    bPlotOutput = args.plotoutput

    # only for debugging
    if __debug__:
        if not cPath:
            cPath = "c:/temp/test.wav"
        if not cNoveltyName:
            cNoveltyName = "Laroche"
        if not bPlotOutput:
            bPlotOutput = True

    # call the function
    computeNoveltyFunctionCl(cPath, cNoveltyName)



# -*- coding: utf-8 -*-
"""
computes the spectral kurtosis from the magnitude spectrum

  Args:
    X: spectrogram (dimension FFTLength X Observations)
    f_s: sample rate of audio data

  Returns:
    vsk spectral kurtosis
"""

import numpy as np
from .FeatureSpectralCentroid import FeatureSpectralCentroid
from .FeatureSpectralSpread import FeatureSpectralSpread


def FeatureSpectralKurtosis(X, f_s, UseBookDefinition=False):

    isSpectrum = X.ndim == 1
    if isSpectrum:
        X = np.expand_dims(X, axis=1)

    if UseBookDefinition:  # not recommended
        # compute mean and standard deviation
        mu_x = np.mean(X, axis=0, keepdims=True)
        std_x = np.std(X, axis=0)

        # remove mean
        X = X - mu_x

        # compute kurtosis
        vsk = np.sum(X**4, axis=0) / (std_x**4 * X.shape[0])
    else:
        f = np.arange(0, X.shape[0]) / (X.shape[0] - 1) * f_s / 2
        # get spectral centroid and spread (mean and std of dist)
        vsc = FeatureSpectralCentroid(X, f_s)  # *2/f_s * (X.shape[0]-1)
        vss = FeatureSpectralSpread(X, f_s)    # *2/f_s * (X.shape[0]-1)

        norm = X.sum(axis=0)
        norm[norm == 0] = 1
        vss[vss == 0] = 1

        # compute kurtosis
        vsk = np.zeros(X.shape[1])
        for n in range(0, X.shape[1]):
            vsk[n] = np.dot((f - vsc[n])**4, X[:, n]) / (vss[n]**4 * norm[n] * X.shape[0])

    return np.squeeze(vsk - 3) if isSpectrum else (vsk - 3)



import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

with open(HERE / "README.md", "r") as fh:
    long_description = fh.read()

setup(name="pyACA",
      use_scm_version=True,
      setup_requires=['setuptools_scm'],
      description="scripts accompanying the book An Introduction to Audio Content Analysis by Alexander Lerch",
      long_description=long_description,
      long_description_content_type="text/markdown",
      classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Topic :: Multimedia :: Sound/Audio :: Analysis",
      ],
      keywords="audio analysis features pitch key extraction music onset beat detection descriptors",
      url="https://github.com/alexanderlerch/pyACA",
      author="Alexander Lerch",
      author_email="info@AudioContentAnalysis.org",
      license="MIT",
      packages=["pyACA"],
      install_requires=[
          "numpy",
          "scipy",
          "matplotlib",
      ],
      zip_safe=False)



# -*- coding: utf-8 -*-
"""
computes the spectral centroid from the (squared) magnitude spectrum

  Args:
    X: spectrogram (dimension FFTLength X Observations)
    f_s: sample rate of audio data

  Returns:
    vsc spectral centroid (in Hz)
"""

import numpy as np


def FeatureSpectralCentroid(X, f_s):

    isSpectrum = X.ndim == 1

    # X = X**2 removed for consistency with book

    norm = X.sum(axis=0, keepdims=True)
    norm[norm == 0] = 1

    vsc = np.dot(np.arange(0, X.shape[0]), X) / norm

    # convert from index to Hz
    vsc = vsc / (X.shape[0] - 1) * f_s / 2

    # if input is a spectrum, output scaler else if spectrogram, output 1d array
    vsc = np.squeeze(vsc) if isSpectrum else np.squeeze(vsc, axis=0)

    return vsc



# -*- coding: utf-8 -*-
"""
NoveltyLaroche
computes the novelty measure used by laroche

  Args:
      X: spectrogram (dimension FFTLength X Observations)
      f_s: sample rate

  Returns:
      d_lar novelty measure

"""

import numpy as np


def NoveltyLaroche(X, f_s):

    # difference spectrum (set first diff to zero)
    X = np.c_[X[:, 0], np.sqrt(X)]

    afDiff = np.diff(X, 1, axis=1)

    # flux
    d_lar = np.sum(afDiff, axis=0) / X.shape[0]

    return (d_lar)



"""
Forms a list of available features by checking for all the Feature*.py files present in the package.

  Args:
    feature_type: (optional) type of features (valid values: 'all', 'spectral' 'temporal')

  Returns:
    features:  list of strings

"""

import glob
import os

import pyACA


def getFeatureList(feature_type='all'):

    feature_type = feature_type.lower()
    if feature_type == 'all':
        feature_type = ''
    elif feature_type == 'spectral':
        feature_type = 'Spectral'
    elif feature_type == 'temporal':
        feature_type = 'Time'
    else:
        print('Invalid feature type')
        return []

    package_loc = os.path.dirname(pyACA.__file__)
    modules = sorted(glob.glob(os.path.join(package_loc, 'Feature' + feature_type + '*.py')))
    features = [os.path.basename(feature)[len('Feature'):-len('.py')] for feature in modules]
    return features



# -*- coding: utf-8 -*-
"""
computeKey

computes a simple beat histogram
  Args:
      afAudioData: array with floating point audio data.
      f_s: sample rate
      afWindow: FFT window of length iBlockLength (default: hann)
      iBlockLength: internal block length (default: 4096 samples)
      iHopLength: internal hop length (default: 2048 samples)

  Returns:
      key string
"""

import numpy as np

from pyACA.computeNoveltyFunction import computeNoveltyFunction
from pyACA.ToolComputeHann import ToolComputeHann
from pyACA.ToolReadAudio import ToolReadAudio

def computeBeatHisto(afAudioData, f_s, afWindow=None, iBlockLength=1024, iHopLength=8):

    # compute window function for FFT
    if afWindow is None:
        afWindow = ToolComputeHann(iBlockLength)

    assert(afWindow.shape[0] == iBlockLength), "parameter error: invalid window dimension"

    # novelty function
    [d, t, peaks] = computeNoveltyFunction('Flux', afAudioData, f_s, afWindow, iBlockLength, iHopLength)

    # compute autocorrelation of result
    afCorr = np.correlate(d, d, "full") / np.dot(d, d)
    afCorr = afCorr[np.arange(d.shape[0], afCorr.size)]

    Bpm = np.flip(60 / t[np.arange(1, t.shape[0])])
    T = np.flip(afCorr)

    return (T, Bpm)


def computeBeatHistoCl(cInPath, cOutPath):
    
    [f_s, afAudioData] = ToolReadAudio(cInPath)

    [T, Bpm] = computeBeatHisto(afAudioData, f_s)

    result = np.vstack((T, Bpm))

    np.savetxt(cOutPath, result)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description='Compute simple beat histogram of wav file')
    parser.add_argument('--infile', metavar='path', required=False,
                        help='path to input audio file')
    parser.add_argument('--outfile', metavar='path', required=False,
                        help='path to output file')

    args = parser.parse_args()
    cInPath = args.infile
    cOutPath = args.outfile

    # only for debugging
    if not cInPath:
        cInPath = "c:/temp/test.wav"
    if not cOutPath:
        cOutPath = "c:/temp/out.txt"

    # call the function
    computeBeatHistoCl(cInPath, cOutPath)



# -*- coding: utf-8 -*-
"""
computes the spectral flux from the magnitude spectrum

  Args:
    X: spectrogram (dimension FFTLength X Observations)
    f_s: sample rate of audio data

  Returns:
    vsf spectral flux
"""

import numpy as np


def FeatureSpectralFlux(X, f_s):

    isSpectrum = X.ndim == 1
    if isSpectrum:
        X = np.expand_dims(X, axis=1)

    # difference spectrum (set first diff to zero)
    X = np.c_[X[:, 0], X]

    afDeltaX = np.diff(X, 1, axis=1)

    # flux
    vsf = np.sqrt((afDeltaX**2).sum(axis=0)) / X.shape[0]

    return np.squeeze(vsf) if isSpectrum else vsf



# -*- coding: utf-8 -*-
"""
computes two peak envelope measures for a time domain signal

  Args:
    x: audio signal
    iBlockLength: block length in samples
    iHopLength: hop length in samples
    f_s: sample rate of audio data (unused)

  Returns:
    vppm peak envelope (1: max, 2: PPM)
    t time stamp

"""

import numpy as np
import pyACA


def FeatureTimePeakEnvelope(x, iBlockLength, iHopLength, f_s):

    # create blocks
    xBlocks = pyACA.ToolBlockAudio(x, iBlockLength, iHopLength)

    # number of results
    iNumOfBlocks = xBlocks.shape[0]

    # compute time stamps
    t = (np.arange(0, iNumOfBlocks) * iHopLength + (iBlockLength / 2)) / f_s

    alpha = 1 - np.array([np.exp(-2.2 / (f_s * 0.01)), np.exp(-2.2 / (f_s * 1.5))])

    # allocate memory
    vppm = np.zeros([2, iNumOfBlocks])
    v_tmp = np.zeros(iBlockLength)

    for n, block in enumerate(xBlocks):
        x_block = np.abs(block)

        # detect the maximum per block
        vppm[0, n] = np.max(x_block)

        # calculate the PPM value - take into account block overlaps
        # and discard concerns wrt efficiency
        v_tmp = ppm(x_block, v_tmp[iHopLength - 1], alpha)
        vppm[1, n] = np.max(v_tmp)

    # convert to dB
    epsilon = 1e-5  # -100dB

    vppm[vppm < epsilon] = epsilon
    vppm = 20 * np.log10(vppm)

    return vppm, t


def ppm(x, filterbuf, alpha):

    # initialization
    ppmout = np.zeros(x.shape[0])

    alpha_AT = alpha[0]
    alpha_RT = alpha[1]

    for i in range(0, x.shape[0]):
        if filterbuf > x[i]:
            # release state
            ppmout[i] = (1 - alpha_RT) * filterbuf
        else:
            # attack state
            ppmout[i] = alpha_AT * x[i] + (1 - alpha_AT) * filterbuf

        filterbuf = ppmout[i]

    return (ppmout)



# -*- coding: utf-8 -*-
"""
helper function: read audio from wav

  Args:
    afAudioData: audio file data
    iBlockLength: processing block length

  Returns:
    afAudioData (array): processed samples
"""

import numpy as np


def ToolPreprocAudio(afAudioData, iBlockLength):

    # pre-processing: downmixing
    if afAudioData.ndim > 1:
        afAudioData = afAudioData.mean(axis=1)
    
    # pre-processing: normalization
    fNorm = np.max(np.abs(afAudioData))
    if fNorm != 0:
        afAudioData = afAudioData / fNorm

    # additional preprocessing step might inlucde sample rate conversion and filtering
    
    # pad with block length zeros just to make sure it runs for weird inputs, too
    afAudioData = np.concatenate((afAudioData, np.zeros([iBlockLength, ])), axis=0)
    
    return (afAudioData)




# -*- coding: utf-8 -*-
"""
computes the ACF coefficients of a time domain signal

  Args:
    x: audio signal
    iBlockLength: block length in samples
    iHopLength: hop length in samples
    f_s: sample rate of audio data (unused)
    eta: index (or vector of indices) of coeff result

  Returns:
    vacf autocorrelation coefficient
    t time stamp
"""

import numpy as np
import pyACA


def FeatureTimeAcfCoeff(x, iBlockLength, iHopLength, f_s, eta=19):

    # create blocks
    xBlocks = pyACA.ToolBlockAudio(x, iBlockLength, iHopLength)

    # number of results
    iNumOfBlocks = xBlocks.shape[0]
    if (np.isscalar(eta)):
        iNumOfResultsPerBlock = 1
    else:
        iNumOfResultsPerBlock = eta.size

    # compute time stamps
    t = (np.arange(0, iNumOfBlocks) * iHopLength + (iBlockLength / 2)) / f_s

    # allocate memory
    vacf = np.zeros([iNumOfResultsPerBlock, iNumOfBlocks])

    for n, block in enumerate(xBlocks):
        # calculate the acf
        if not block.sum():
            vacf[np.arange(0, iNumOfResultsPerBlock), n] = np.zeros(iNumOfResultsPerBlock)
            continue
        else:
            afCorr = np.correlate(block, block, "full") / np.dot(block, block)

        # find the coefficients specified in eta
        vacf[np.arange(0, iNumOfResultsPerBlock), n] = afCorr[iBlockLength + eta]

    return vacf, t



# -*- coding: utf-8 -*-
"""
computes the spectral slope from the magnitude spectrum

  Args:
    X: spectrogram (dimension FFTLength X Observations)
    f_s: sample rate of audio data

  Returns:
    vssl spectral slope
"""

import numpy as np


def FeatureSpectralSlope(X, f_s):

    # compute mean
    mu_x = X.mean(axis=0, keepdims=True)

    # compute index vector
    kmu = np.arange(0, X.shape[0]) - X.shape[0] / 2

    # compute slope
    X = X - mu_x
    vssl = np.dot(kmu, X) / np.dot(kmu, kmu)

    return vssl



# -*- coding: utf-8 -*-
"""
computePitch

computes the fundamental frequency of the (monophonic) audio
supported pitch trackers are:
    'SpectralAcf',
    'SpectralHps',
    'TimeAcf',
    'TimeAmdf',
    'TimeAuditory',
    'TimeZeroCrossings',
  Args:
      cPitchTrackName: feature to compute, e.g. 'SpectralHps'
      afAudioData: array with floating point audio data.
      f_s: sample rate
      afWindow: FFT window of length iBlockLength (default: hann)
      iBlockLength: internal block length (default: 4096 samples)
      iHopLength: internal hop length (default: 2048 samples)

  Returns:
      f frequency
      t time stamp for the frequency value
"""

import numpy as np
from scipy.signal import spectrogram
import matplotlib.pyplot as plt

import pyACA
from pyACA.ToolPreprocAudio import ToolPreprocAudio
from pyACA.ToolComputeHann import ToolComputeHann
from pyACA.ToolReadAudio import ToolReadAudio


def computePitch(cPitchTrackName, afAudioData, f_s, afWindow=None, iBlockLength=4096, iHopLength=2048):
    
    #mypackage = __import__(".Pitch" + cPitchTrackName, package="pyACA")
    hPitchFunc = getattr(pyACA, "Pitch" + cPitchTrackName)

    # pre-processing
    afAudioData = ToolPreprocAudio(afAudioData, iBlockLength)

    if isSpectral(cPitchTrackName):
        # compute window function for FFT
        if afWindow is None:
            afWindow = ToolComputeHann(iBlockLength)

        assert(afWindow.shape[0] == iBlockLength), "parameter error: invalid window dimension"

        # in the real world, we would do this block by block...
        [f_k, t, X] = spectrogram(afAudioData,
                                  f_s,
                                  afWindow,
                                  iBlockLength,
                                  iBlockLength - iHopLength,
                                  iBlockLength,
                                  False,
                                  True,
                                  'spectrum')

        # we just want the magnitude spectrum...
        X = np.sqrt(X / 2)

        # compute instantaneous pitch chroma
        f = hPitchFunc(X, f_s)

    if isTemporal(cPitchTrackName):
        [f, t] = hPitchFunc(afAudioData, iBlockLength, iHopLength, f_s)

    return (f, t)


#######################################################
# helper functions
def isSpectral(cName):
    bResult = False
    if "Spectral" in cName:
        bResult = True

    return (bResult)


def isTemporal(cName):
    bResult = False
    if "Time" in cName:
        bResult = True

    return (bResult)


def computePitchCl(cPath, cPitchTrackName, bPlotOutput=False):
    
    # read audio file
    [f_s, afAudioData] = ToolReadAudio(cPath)
    # afAudioData = np.sin(2*np.pi * np.arange(f_s*1)*440./f_s)

    # compute feature
    [v, t] = computePitch(cPitchTrackName, afAudioData, f_s)

    # plot feature output
    if bPlotOutput:
        plt.plot(t, v)

    return (v, t)


if __name__ == "__main__":
    import argparse

    # add command line args and parse them
    parser = argparse.ArgumentParser(description='Compute key of wav file')
    parser.add_argument('--infile', metavar='path', required=False,
                        help='path to input audio file')
    parser.add_argument('--featurename', metavar='string', required=False,
                        help='feature name in the format SpectralPitchChroma')
    parser.add_argument('--plotoutput', metavar='bool', required=False,
                        help='option to plot the output')

    # retrieve command line args
    args = parser.parse_args()
    cPath = args.infile
    cPitchTrackName = args.featurename
    bPlotOutput = args.plotoutput

    # only for debugging
    if __debug__:
        if not cPath:
            cPath = "c:/temp/test.wav"
        if not cPitchTrackName:
            cPitchTrackName = "TimeAmdf"
        if not bPlotOutput:
            bPlotOutput = True

    # call the function
    computePitchCl(cPath, cPitchTrackName, bPlotOutput)



# -*- coding: utf-8 -*-
"""
computes the maximum of the spectral autocorrelation function

  Args:
    X: spectrogram (dimension FFTLength X Observations)
    f_s: sample rate of audio data

  Returns:
    f acf maximum location (in Hz)
"""

import numpy as np
from scipy.signal import find_peaks


def PitchSpectralAcf(X, f_s):

    # initialize
    f_min = 300
    f = np.zeros(X.shape[1])

    # use spectral symmetry for robustness
    X[0, :] = np.max(X)
    X = np.concatenate((np.flipud(X), X), axis=0)

    # compute the ACF
    for n in range(0, X.shape[1]):

        if X[:, n].sum() < 1e-20:
            continue

        eta_min = int(round(f_min / f_s * (X.shape[0] - 2))) - 1

        afCorr = np.correlate(X[:, n], X[:, n], "full") / np.dot(X[:, n], X[:, n])
        afCorr = afCorr[np.arange(X.shape[0], afCorr.size)]

        # find the highest local maximum
        iPeaks = find_peaks(afCorr, height=0)
        if iPeaks[0].size:
            eta_min = np.max([eta_min, iPeaks[0][0] - 1])
        f[n] = np.argmax(afCorr[np.arange(eta_min, afCorr.size)]) + 1

        # find max index and convert to Hz (note: X has double length)
        f[n] = (f[n] + eta_min) / (X.shape[0] - 2) * f_s

    return (f)



"""
Runs MATLAB scripts through python and cross checks results of python modules with corresponding MATLAB code.

Requirements:
    - matlab.engine must be installed. (https://www.mathworks.com/help/matlab/matlab_external/install-the-matlab-engine-for-python.html)
    - The engine requires python 3.6 or less.
    - Need to specify path to MATLAB code.

More info:
    - https://www.mathworks.com/help/matlab/matlab-engine-for-python.html
    - https://www.mathworks.com/matlabcentral/answers/415682-how-to-run-m-file-in-python

"""


from importlib import util as imputil
matlab_spec = imputil.find_spec("matlab")  # importlib.util.find_spec works for python3 >= v3.4
if matlab_spec is not None:
    import matlab.engine

import os
import sys
import logging
import unittest
import numpy as np
import numpy.testing as npt

import pyACA

# Initialize this according to the path on your system:
path_to_matlab_code = '/Users/kaushal/Kaushal/Repositories/ACA-Code'


class TestFeaturesWithMatlab(unittest.TestCase):
    logger = logging.getLogger(__name__)
    logging.basicConfig(level=logging.INFO)

    def __init__(self, *args, **kwargs):
        super(TestFeaturesWithMatlab, self).__init__(*args, **kwargs)
        self.matlab_engine = None
        path_valid = os.path.exists(path_to_matlab_code)
        init_engine = True

        if 'matlab.engine' not in sys.modules:
            self.logger.warning("matlab package not found")
            init_engine = False
        if not path_valid:
            self.logger.warning("Invalid path to Matlab code")
            init_engine = False

        if init_engine:
            self.matlab_engine = matlab.engine.start_matlab()
            self.matlab_engine.cd(path_to_matlab_code)

    def __del__(self):
        if self.matlab_engine is not None:
            self.matlab_engine.quit()

    def test_all_features(self):

        if self.matlab_engine is None:
            self.skipTest("Matlab engine not available")

        # WhiteNoise input
        fs = 44100
        x_py = np.random.uniform(-1, 1, size=(fs//2, 1))  # 0.5 sec
        x_m = matlab.double(x_py.tolist())

        for feature in pyACA.getFeatureList():
            self.logger.info('Testing feature:' + feature)
            v_out_py, t_py = pyACA.computeFeature(feature, x_py, fs)
            # Note: fs must be float when passing to Matlab
            v_out_m, t_m = self.matlab_engine.ComputeFeature(feature, x_m, float(fs), nargout=2)

            v_out_py = v_out_py.squeeze()
            v_out_m = np.asarray(v_out_m).squeeze()

            with self.subTest(msg=feature):
                npt.assert_almost_equal(v_out_py, v_out_m, decimal=7, err_msg="MATLAB crosscheck test failed for " + feature)


